/*
        Copyright (c) 2006 Martin Kaltenbrunner <mkalten@iua.upf.es>
                  (c) 2007 Andrew Walkingshaw <adw27@cam.ac.uk>
                  
        Permission is hereby granted, free of charge, to any person obtaining
        a copy of this software and associated documentation files
        (the "Software"), to deal in the Software without restriction,
        including without limitation the rights to use, copy, modify, merge,
        publish, distribute, sublicense, and/or sell copies of the Software,
        and to permit persons to whom the Software is furnished to do so,
        subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
        EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
        MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
        IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR
        ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
        CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
        WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

import tuio.*;


Molecule mol;
TuiCursorList cursorList;
TuioClient client;
MoietyDict moieties;
AdjacencyMatrix adj;

void setup()
{
  size(640,480);
  noStroke();
  fill(0);
  noLoop();
  redraw();
  mol = new Molecule();
  cursorList = new TuiCursorList();
  client  = new TuioClient(this);
  moieties = new MoietyDict();
  moieties.load_moieties("moieties.txt");
}

void draw()
{
  background(255);
  cursorList.draw();
  mol.draw();
}

// called after each message bundle
void refresh() { redraw(); }

// called when an object is added to the scene
void addTuiObject(Integer s_id, Integer f_id) {
  mol.activate(s_id,f_id);
  //System.out.println("add object "+f_id+" ("+s_id+")");
}

// called when an object is removed from the scene
void removeTuiObject(Integer s_id,Integer f_id ) {
  mol.deactivate(s_id);
  //System.out.println("remove object "+f_id+" ("+s_id+")");
}

// called when an object is moved
void updateTuiObject (Integer s_id, Integer f_id, Float xpos, Float ypos, Float angle) {
  mol.update(s_id,(int)(640*xpos.floatValue()),(int)(480*ypos.floatValue()),angle.floatValue());
  //System.out.println("update object "+f_id+" ("+s_id+")");
}

// called when a cursor appears the scene
void addTuiCursor(Integer s_id) {
  //System.out.println("add cursor "+s_id);
  cursorList.add(s_id);
  redraw();
}

// called when a cursor is removed from the scene
void removeTuiCursor(Integer s_id) {
  //System.out.println("remove cursor "+s_id);
  cursorList.remove(s_id);
  redraw();
}

// called when a cursor is moved
void updateTuiCursor (Integer s_id, Float xpos, Float ypos) {
  //System.out.println("update cursor "+s_id+" "+xpos+" "+ypos);
  cursorList.update(s_id,(int)(640*xpos.floatValue()),(int)(480*ypos.floatValue()));
  redraw();
}
//--------------------------------------------------------------------------------------------------
//helper classes

// draw SMILES when the mouse is clicked
void mousePressed() {
  // need to unset smilesed before we check for this
  // to work...
  Enumeration e = mol.get_atomList().elements();
  while (e.hasMoreElements()) {
    Atom obj = ((Atom)e.nextElement());
    obj.unsetSMILES();
  }
  String smiles = mol.getSMILES();
  link("http://pubchem.ncbi.nlm.nih.gov/search/search.cgi?smarts=" + smiles);
}

class AdjacencyMatrix {
  java.util.Hashtable adjacency;
  
  AdjacencyMatrix() {
    adjacency = new Hashtable();
  }

  void add_atom(Atom a) {
    adjacency.put(a, new java.util.Hashtable()); 
  }  
  
  void set_adjacency(Atom a, Atom b) {
    java.util.Hashtable x = (java.util.Hashtable)adjacency.get(a);
    java.util.Hashtable y = (java.util.Hashtable)adjacency.get(b);
    x.put(b, new Boolean(true));
    y.put(a, new Boolean(true));
  }
  
  int bondcount(Atom a) {
    java.util.Hashtable t = (java.util.Hashtable)adjacency.get(a);
    return t.size();
  }
  
  Boolean is_adjacent(Atom a, Atom b) {
    if (adjacency.containsKey(a)) {
      java.util.Hashtable atadj = (java.util.Hashtable)adjacency.get(a);
      if (atadj.containsKey(b)) {
        return (Boolean)atadj.get(b);
      } 
    }
    return new Boolean(false);
  }
}

class Moiety {
  String smiles;
  int valency;
  String name;
  color col;
  color txt;
  
  Moiety (String n, int v, String s, int colr, int colg, int colb, int txtr, int txtg, int txtb) {
    name = n;
    valency = v;
    smiles = s;
    col = color(colr, colg, colb);
    txt = color(txtr, txtg, txtb);
  }
  
  color get_color() {
   return col;
  }
  
  color get_textcolor() {
    return txt;
  }
  
  String get_name() {
    return name;
  }
  
  int get_valency() {
    return valency;
  }
  
 String get_smiles() {
    return smiles;
  }
}

  
class MoietyDict {
  java.util.Hashtable Moieties;
  
  MoietyDict() {
    Moieties = new Hashtable();
  }

  void add_moiety(String name, int valence, String smiles, Integer fiducial_id, int colr, int colg, int colb, int txtr, int txtg, int txtb) {
    Moiety m = new Moiety(name, valence, smiles, colr, colg, colb, txtr, txtg, txtb);
    Moieties.put(fiducial_id, m); 
  }
  
  Moiety get_moiety(Integer f_id) {
    return (Moiety)Moieties.get(f_id);
  }
  
  void load_moieties(String filename) {
    String[] lines = loadStrings(filename);
    int index=0;
    while (index < lines.length) {
      String[] pieces = split(lines[index], ',');
      if (pieces.length == 10) {
        add_moiety(pieces[1], int(pieces[2]), pieces[3], new Integer(pieces[0]), int(pieces[4]), int(pieces[5]), int(pieces[6]), int(pieces[7]), int(pieces[8]), int(pieces[9]));
      }
      index++;
    }
  }
}

class Molecule {

  java.util.Hashtable atomList;
// CMLMolecule cml_mol;
  Molecule() {
    textFont(createFont("Sans", 12));
    atomList = new Hashtable();
  }

  Hashtable get_atomList() {
    return atomList;
  }
  
//  void calc_CMLMolecule() {
//    cml_mol = new CMLMolecule();
//    from.setId("ChemReact");
//    Enumeration e = atomList.elements();
//    while (e.hasMoreElements()) {
//      Atom this_atom = ((Atom)e.nextElement());
//      CMLAtom a = new CMLAtom();
//      a.setElementType(a.get_atomType());
//      a.setId(a.get_session_id());
//      cml_mol.addAtom(a);
//    } // atoms now added to document
//  }
  
  void draw() { 
    Enumeration e = atomList.elements();
    while (e.hasMoreElements()) {
      Atom obj = ((Atom)e.nextElement());
      int posx = obj.getx();
      int posy = obj.gety();
      Vector adjacencyList = obj.get_adjacencyList();
      Enumeration ae = adjacencyList.elements();
      while (ae.hasMoreElements()) {
        boolean multiple = false;
        // draw bonds
        stroke(color(0));
        Integer sid = (Integer)ae.nextElement();
        Atom adjacentObject = getObject(sid);
        // are we double-bonded with this branch?
        if ((! obj.saturated()) && (! adjacentObject.saturated())) {
          multiple = true;
        }
        int posx_other = adjacentObject.getx();
        int posy_other = adjacentObject.gety();
        int deltax = 0;
        int deltay = 0;
        if (multiple) { // double bond
          int xdiff = abs(posx_other - posx);
          int ydiff = abs(posy_other - posy);
          if (xdiff > ydiff) {
            line(posx, posy+10, posx_other, posy_other+10);
            deltay = -5;
          } else {
            line(posx+25, posy, posx_other+25, posy_other);
            deltax = -5;
          }
        }      
        line(posx+deltax,posy+deltay,
             posx_other+deltax,posy_other+deltay); 
      }
      obj.draw();
    }
    String smiles = mol.getSMILES();
    fill(0); stroke(0);
    text(smiles, 20, 20);  
  }
  
  String getSMILES_atom(Atom atom) {
    String smiles = "";

    String atomtype = atom.get_smiles_string();
    // add this atom to the SMILES string
    atom.setSMILES();
    if (atomtype != "H") {
      smiles += atomtype;
      Vector adjacencyList = atom.get_adjacencyList();
      Enumeration ae = adjacencyList.elements();
    
      while (ae.hasMoreElements()) {
        boolean multiple = false;
        Integer sid = (Integer)ae.nextElement();
        Atom adjacent_atom = getObject(sid);
        // recurse on the ones we haven't visited
        if (! adjacent_atom.SMILES_status() &&
             (adjacent_atom.get_smiles_string() != "H")) {
          smiles += "(";
          // are we double-bonded with this branch?
          if ((! atom.saturated()) && (! adjacent_atom.saturated())) {
            smiles += "=";
          }
          smiles += getSMILES_atom(adjacent_atom);
          // close the branch
          smiles += ")";
        }
      }
    }
    return smiles;  
  }
  
  String getSMILES() { 
    String smiles = "";
    Enumeration e = atomList.elements();
    while (e.hasMoreElements()) {
      String this_smiles = "";
      Atom atom = ((Atom)e.nextElement());
      if (! atom.SMILES_status()) {
        smiles += getSMILES_atom(atom);
      }
    }
    return smiles;
  }
  
  Atom getObject(Integer s_id) {
    return (Atom)atomList.get(s_id);
  }
  
  void activate(Integer s_id, Integer f_id) { 
    Atom nobj = new Atom(s_id.intValue(), f_id.intValue(), moieties);
    atomList.put(s_id,nobj);
  }
  
  void deactivate(Integer s_id) {
    atomList.remove(s_id);
  }
  
  void update(Integer s_id, int x, int y, float a) {
    Atom tobj = (Atom)atomList.get(s_id);
    if (tobj!=null) {
        java.util.Vector adjacencyList = new Vector();
        java.util.Hashtable tmpadj = new Hashtable();
        Enumeration e = atomList.elements();
        while (e.hasMoreElements()) {
          java.util.Hashtable adjdist;
          Atom nextel = ((Atom)e.nextElement());
          int id = nextel.get_session_id();
          int xpos = nextel.getx(); 
          int ypos = nextel.gety();
          float delta = pow(pow((x-xpos),2) + pow((y-ypos),2), 0.5);
          if (s_id.intValue() != id && delta < 120) {
            tmpadj.put(new Float(delta), new Integer(id));
          }
        }
        TreeSet distSet = new TreeSet(tmpadj.keySet());
        Iterator it = distSet.iterator();
        while ((adjacencyList.size() < tobj.get_valency()) &&
                it.hasNext() ) {
          Float k = (Float)it.next();
          Integer index = (Integer)tmpadj.get(k);
          Atom other_atom = (Atom)atomList.get(index);
          if (other_atom.get_adjacencyList().contains(s_id)) {
            adjacencyList.add(index);
          } 
          else if (other_atom.get_adjacencyList().size() < 
            other_atom.get_valency()) 
          {
//            println("adjusting valency of " + index + 
//              " (was: " + other_atom.get_adjacencyList().size() + " of " +
//              other_atom.get_valency()  +" )");
            adjacencyList.add(index);
            Vector al = other_atom.get_adjacencyList();
            al.add(s_id);
            other_atom.set_adjacencyList(al);
          }
        }
        tobj.update(x,y,a,adjacencyList);
    }
  }
}

class Atom {
  int session_id, fiducial_id;
  int valency;
  int xpos, ypos;
  float angle;
  color col;
  color txtcol;
  java.util.Vector adjacencyList;
  String name;
  String smiles_string;
  boolean smilesed = false;
  
  Atom(int s_id,int f_id, MoietyDict m) {
    session_id  = s_id; fiducial_id = f_id;
    Moiety moi = m.get_moiety(new Integer(f_id));
    name = moi.get_name();
    smiles_string = moi.get_smiles();
    valency = moi.get_valency();
    col = moi.get_color();
    txtcol = moi.get_textcolor(); 
    xpos = 0; 
    ypos = 0; 
    angle = 0.0f;
    adjacencyList = new Vector();
  }
  
  int getx() {
    return xpos;
  }
  
  int gety() {
    return ypos;
  }

  String get_smiles_string() {
    return smiles_string;
  }
  
  void setSMILES() {
    smilesed = true;
  }
  
  void unsetSMILES() {
    smilesed = false;
  }
  
  boolean SMILES_status() {
    return smilesed;
  }
  
  int get_session_id() {
    return session_id;
  }
  
  void set_adjacencyList(Vector al) {
    adjacencyList = al;
  }
  
  Vector get_adjacencyList() {
    return adjacencyList;
  }
  
  int get_valency() {
    return valency;
  }
  
  boolean saturated() {
   int valency_theoretical = get_valency();
   int valency_actual = get_adjacencyList().size();
   if (valency_theoretical != valency_actual) {
     return false;
   }
   return true;
  }
  
  void update(int x, int y, float a, java.util.Vector l) {
    xpos=x;
    ypos=y;
    angle=a;
    adjacencyList = l;
    smilesed = false;
  }

  void draw() {
    // set up colors
   colorMode(RGB, 255);
   color warning = color(0,255,0);
   stroke(color(0,0,0));   
   fill(color(255,255,255));
   String tx = name;
      
  if (!saturated()) {
     stroke(warning);
     txtcol = warning;
   }
   
   fill(col);
   pushMatrix();
   translate(xpos,ypos);
   ellipse(0,0,50,50);
   popMatrix();
   fill(txtcol);
   text(""+tx, xpos, ypos);
  }
}
  class TuiCursorList {

  java.util.Hashtable cursorList;

  TuiCursorList() {
    cursorList = new Hashtable();
  }

  void draw() {
    stroke(0,0,255);
    Enumeration e = cursorList.elements();
    while (e.hasMoreElements()) {
      Vector pointList = (Vector)e.nextElement();
      
      if (pointList.size()>0) {
        Point start_point = (Point)pointList.elementAt(0);
        for (int i=0;i<pointList.size();i++) {
           Point end_point = (Point)pointList.elementAt(i);
           line(start_point.x,start_point.y,end_point.x,end_point.y);
           start_point = end_point;
        } 
      }
    }
  }
  
  void add(Integer s_id) { 
    Vector pointList = new Vector();
    cursorList.put(s_id,pointList);
  }
  
  void remove(Integer s_id) {
    cursorList.remove(s_id);
  }
  
  void update(Integer s_id, int xpos, int ypos) {
    Vector pointList = (Vector)cursorList.get(s_id);
    if (pointList!=null) pointList.addElement(new Point(xpos,ypos));
  }
}
